/*
 * Copyright (c) 2016 Intel Corporation.
 * Copyright (c) 2019-2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr.h>
#include <logging/log.h>
#include <usb/usb_device.h>
#include <fs/fs.h>
#include <stdio.h>

#include <drivers/counter.h>

LOG_MODULE_REGISTER(main);

#if CONFIG_DISK_DRIVER_FLASH
#include <storage/flash_map.h>
#endif

#if CONFIG_FAT_FILESYSTEM_ELM
#include <ff.h>
#endif

#if CONFIG_FILE_SYSTEM_LITTLEFS
#include <fs/littlefs.h>
FS_LITTLEFS_DECLARE_DEFAULT_CONFIG(storage);
#endif

#define COUNTER_DEVICE DT_LABEL(DT_NODELABEL(rtc0))

const struct device *counter_dev;

static struct fs_mount_t fs_mnt;
struct fs_file_t file;
char fname[255];

static uint64_t write_to_file(struct fs_file_t file, uint8_t *buf, size_t buf_len) {
    counter_dev = device_get_binding(COUNTER_DEVICE);
    if (!counter_dev) {
        printk("Counter non-existant?\n");
    }
    
    int rc = counter_start(counter_dev);
    if (rc) {
        printk("Failed to start counter: %d", rc);
    }

    fs_file_t_init(&file);


    rc = fs_open(&file, fname, FS_O_CREATE | FS_O_RDWR | FS_O_APPEND);
    if (rc < 0) {
        printk("FAIL: open %s: %d\n", fname, rc);
    }

    uint32_t start_ticks = 0;
    uint32_t stop_ticks = 0;

    counter_get_value(counter_dev, &start_ticks);
    rc = fs_write(&file, buf, buf_len);
    counter_get_value(counter_dev, &stop_ticks);

    if (rc != buf_len) {
        printk("Failed to write file %d\n", rc);
    } else {
        printk("Time to write %d bytes to file: %lld us\n", buf_len, counter_ticks_to_us(counter_dev, stop_ticks - start_ticks));
    }

    rc = fs_close(&file);
    //printk("%s close: %d\n", fname, rc);

    return counter_ticks_to_us(counter_dev, stop_ticks - start_ticks);
}

static int setup_flash(struct fs_mount_t *mnt)
{
	int rc = 0;
#if CONFIG_DISK_DRIVER_FLASH
	unsigned int id;
	const struct flash_area *pfa;

	mnt->storage_dev = (void *)FLASH_AREA_ID(fstorage);
	id = (uintptr_t)mnt->storage_dev;

	rc = flash_area_open(id, &pfa);
	printk("Area %u at 0x%x on %s for %u bytes\n",
	       id, (unsigned int)pfa->fa_off, pfa->fa_dev_name,
	       (unsigned int)pfa->fa_size);

	if (rc < 0 && IS_ENABLED(CONFIG_APP_WIPE_STORAGE)) {
		printk("Erasing flash area ... ");
		rc = flash_area_erase(pfa, 0, pfa->fa_size);
		printk("%d\n", rc);
	}

	if (rc < 0) {
		flash_area_close(pfa);
	}
#endif
	return rc;
}

static int mount_app_fs(struct fs_mount_t *mnt)
{
	int rc;

#if CONFIG_FAT_FILESYSTEM_ELM
	static FATFS fat_fs;

	mnt->type = FS_FATFS;
	mnt->fs_data = &fat_fs;
	if (IS_ENABLED(CONFIG_DISK_DRIVER_RAM)) {
		mnt->mnt_point = "/RAM:";
	} else if (IS_ENABLED(CONFIG_DISK_DRIVER_SDMMC)) {
		mnt->mnt_point = "/SD:";
	} else {
		mnt->mnt_point = "/NAND:";
	}

#elif CONFIG_FILE_SYSTEM_LITTLEFS
	mnt->type = FS_LITTLEFS;
	mnt->mnt_point = "/lfs";
	mnt->fs_data = &storage;
#endif
	rc = fs_mount(mnt);

	return rc;
}

static void setup_disk(void)
{
	struct fs_mount_t *mp = &fs_mnt;
	struct fs_dir_t dir;
	struct fs_statvfs sbuf;
	int rc;

	fs_dir_t_init(&dir);

	if (IS_ENABLED(CONFIG_DISK_DRIVER_FLASH)) {
		rc = setup_flash(mp);
		if (rc < 0) {
			LOG_ERR("Failed to setup flash area");
			return;
		}
	}

	if (!IS_ENABLED(CONFIG_FILE_SYSTEM_LITTLEFS) &&
	    !IS_ENABLED(CONFIG_FAT_FILESYSTEM_ELM)) {
		LOG_INF("No file system selected");
		return;
	}

	rc = mount_app_fs(mp);
	if (rc < 0) {
		LOG_ERR("Failed to mount filesystem");
		return;
	}

	/* Allow log messages to flush to avoid interleaved output */
	k_sleep(K_MSEC(50));

	printk("Mounted %s\n", fs_mnt.mnt_point);

	rc = fs_statvfs(mp->mnt_point, &sbuf);
	if (rc < 0) {
		printk("FAIL: statvfs: %d\n", rc);
		return;
	}

	printk("%s: bsize = %lu ; frsize = %lu ; blocks = %lu ; bfree = %lu\n",
        mp->mnt_point,
        sbuf.f_bsize,
        sbuf.f_frsize,
        sbuf.f_blocks,
        sbuf.f_bfree
    );

#if (IS_ENABLED(CONFIG_APP_UNLINK_WRITE_TEST_FILE))
    printk("Deleting file...");
    snprintf(fname, sizeof(fname), "%s/test_file.txt", mp->mnt_point);
    rc = fs_unlink(fname);
    if (rc) {
        printk("deleting file err: %d\n", rc);
    }
    printk("DONE\n");

    int file_write_size_kb = 48;
    uint8_t fake_data[1024];
    memset(fake_data, 'a', sizeof(fake_data));
    uint64_t total_ticks = 0;
    for (uint32_t i = 0; i < file_write_size_kb; i += 1) {
        printk("%d : ", i);
        total_ticks += write_to_file(file, fake_data, (size_t)1024);
    }

    printk("Took total_ticks_us to write %dKB: %llu\n", file_write_size_kb, total_ticks);
#endif // IS_ENABLED(CONFIG_APP_UNLINK_WRITE_TEST_FILE)

	rc = fs_opendir(&dir, mp->mnt_point);
	printk("%s opendir: %d\n", mp->mnt_point, rc);

	if (rc < 0) {
		LOG_ERR("Failed to open directory");
	}

	while (rc >= 0) {
		struct fs_dirent ent = { 0 };

		rc = fs_readdir(&dir, &ent);
		if (rc < 0) {
			LOG_ERR("Failed to read directory entries");
			break;
		}
		if (ent.name[0] == 0) {
			printk("End of files\n");
			break;
		}
		printk("  %c %u %s\n",
		       (ent.type == FS_DIR_ENTRY_FILE) ? 'F' : 'D',
		       ent.size,
		       ent.name);
	}

	(void)fs_closedir(&dir);

	return;
}

void main(void)
{
	int ret;

	setup_disk();

	ret = usb_enable(NULL);
	if (ret != 0) {
		LOG_ERR("Failed to enable USB");
		return;
	}

	LOG_INF("The device is put in USB mass storage mode.\n");
}
